# Usage: ls-ingest <dataset_id>
import os
import json
import argparse
import pandas as pd

from latentscope.util import get_data_dir

# TODO make a parquet version of these
def main():
    parser = argparse.ArgumentParser(description='Ingest a dataset')
    parser.add_argument('id', type=str, help='Dataset id (directory name in data folder)')
    parser.add_argument('--path', type=str, help='Path to csv or parquet file, otherwise assumes input.csv in dataset directory')
    args = parser.parse_args()
    ingest_file(args.id, args.path)

def ingest_file(dataset_id, file_path):
    DATA_DIR = get_data_dir()
    directory = os.path.join(DATA_DIR, dataset_id)
    if not file_path:
        file_path = os.path.join(directory, "input.csv")
    file_type = file_path.split('.')[-1]
    print(f"File type detected: {file_type}")
    file = os.path.join(file_path)
    print("reading", file)
    if file_type == "csv":
        df = pd.read_csv(file)
    elif file_type == "parquet":
        df = pd.read_parquet(file)
    elif file_type == "jsonl":
        with open(file, 'r') as f:
            lines = f.readlines()
            df = pd.DataFrame([json.loads(line) for line in lines])
    else:
        raise ValueError(f"Unsupported file type: {file_type}")
    ingest(dataset_id, df)


def ingest(dataset_id, df, text_column = None):
    DATA_DIR = get_data_dir()
    print("DATA DIR", DATA_DIR)
    directory = os.path.join(DATA_DIR, dataset_id)
    print("DIRECTORY", directory)
    if not os.path.exists(directory):
        os.makedirs(directory)
    df = df.reset_index(drop=True)
    print(df.head())
    print(df.tail())
    print(df.columns)
    output_file = f"{directory}/input.parquet"
    df.to_parquet(output_file)
    print("wrote", output_file)
    # write out a json file with the model name and shape of the embeddings
    if text_column is None:
        text_column = "text" in df.columns and "text" or df.columns[0]
    with open(os.path.join(directory,'meta.json'), 'w') as f:
        json.dump({
            "id": dataset_id,
            "length": df.shape[0],
            "columns": df.columns.tolist(),
            "text_column": text_column,
            }, f, indent=2)

    # create all the directories we will use
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "tags"), exist_ok=True)
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "embeddings"), exist_ok=True)
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "umaps"), exist_ok=True)
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "clusters"), exist_ok=True)
    os.makedirs(os.path.join(DATA_DIR, dataset_id, "scopes"), exist_ok=True)


if __name__ == "__main__":
    main()



# Usage: ls-label <dataset_id> <text_column> <cluster_id> <model_id> <context>
import os
import re
import sys
import json
import time
import argparse
import numpy as np
import pandas as pd
try:
    # Check if the runtime environment is a Jupyter notebook
    if 'ipykernel' in sys.modules and 'IPython' in sys.modules:
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm
except ImportError as e:
    # Fallback to the standard console version if import fails
    from tqdm import tqdm

from latentscope.util import get_data_dir
from latentscope.models import get_chat_model

def chunked_iterable(iterable, size):
    """Yield successive chunks from an iterable."""
    for i in range(0, len(iterable), size):
        yield iterable[i:i + size]

def too_many_duplicates(line, threshold=10):
    word_count = {}
    if line is None:
        return False
    words = line.split()
    for word in words:
        word_count[word] = word_count.get(word, 0) + 1
    return any(count > threshold for count in word_count.values())

def main():
    parser = argparse.ArgumentParser(description='Label a set of slides using OpenAI')
    parser.add_argument('dataset_id', type=str, help='Dataset ID (directory name in data/)')
    parser.add_argument('text_column', type=str, help='Output file', default='text')
    parser.add_argument('cluster_id', type=str, help='ID of cluster set', default='cluster-001')
    parser.add_argument('model_id', type=str, help='ID of model to use', default="openai-gpt-3.5-turbo")
    parser.add_argument('context', type=str, help='Additional context for labeling model', default="")
    parser.add_argument('--rerun', type=str, help='Rerun the given embedding from last completed batch')

    # Parse arguments
    args = parser.parse_args()

    labeler(args.dataset_id, args.text_column, args.cluster_id, args.model_id, args.context, args.rerun)


def labeler(dataset_id, text_column="text", cluster_id="cluster-001", model_id="openai-gpt-3.5-turbo", context="", rerun=""):
    DATA_DIR = get_data_dir()
    df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "input.parquet"))

    # Load the indices for each cluster from the prepopulated labels file generated by cluster.py
    cluster_dir = os.path.join(DATA_DIR, dataset_id, "clusters")
    clusters = pd.read_parquet(os.path.join(cluster_dir, f"{cluster_id}-labels-default.parquet"))
    # initialize the labeled property to false when loading default clusters
    clusters = clusters.copy()
    clusters['labeled'] = False

    unlabeled_row = 0
    if rerun is not None:
        label_id = rerun
        clusters = pd.read_parquet(os.path.join(cluster_dir, f"{label_id}.parquet"))
        # print(clusters.columns)
        # find the first row where labeled isnt True
        unlabeled_row = clusters[~clusters['labeled']].first_valid_index()
        print(f"First unlabeled row: {unlabeled_row}")
        

    else:
        # Determine the label id for the given cluster_id by checking existing label files
        label_files = [f for f in os.listdir(cluster_dir) if re.match(rf"{re.escape(cluster_id)}-labels-\d+\.parquet", f)]
        if label_files:
            # Extract label numbers and find the maximum
            label_numbers = [int(re.search(rf"{re.escape(cluster_id)}-labels-(\d+)\.parquet", f).group(1)) for f in label_files]
            next_label_number = max(label_numbers) + 1
        else:
            next_label_number = 1
        label_id = f"{cluster_id}-labels-{next_label_number:03d}"
    print("RUNNING:", label_id)

    model = get_chat_model(model_id)
    model.load_model()
    enc = model.encoder

    system_prompt = {"role":"system", "content": f"""You're job is to summarize lists of items with a short label of no more than 4 words. 
{context}
The user will submit a bulleted list of items and you should choose a label that best summarizes the theme of the list so that someone browsing the labels will have a good idea of what is in the list. 
Do not use punctuation, just return a few words that summarize the list."""}

    # TODO: why the extra 10 for openai?
    max_tokens = model.params["max_tokens"] - len(enc.encode(system_prompt["content"])) - 10

    # Create the lists of items we will send for summarization
    # Current looks like:
    # 1. item 1
    # 2. item 2
    # ...
    # we truncate the list based on tokens and we also remove items that have too many duplicate words
    extracts = []
    for _, row in clusters.iterrows():
        indices = row['indices']
        items = df.loc[list(indices), text_column]
        items = items.drop_duplicates()
        text = '\n'.join([f"{i+1}. {t}" for i, t in enumerate(items) if not too_many_duplicates(t)])
        # print(text)
        encoded_text = enc.encode(text)
        if len(encoded_text) > max_tokens:
            # print("truncating", len(encoded_text), "tokens")
            encoded_text = encoded_text[:max_tokens]
            # print("truncated to", len(encoded_text), "tokens")
        extracts.append(enc.decode(encoded_text))

    # TODO we arent really batching these
    batch_size = 1
    labels = []
    clean_labels = []


    for i,batch in enumerate(tqdm(chunked_iterable(extracts, batch_size),  total=len(extracts)//batch_size)):
        # print(batch[0])
        if(unlabeled_row > 0):
            if clusters.loc[i, 'labeled']:
                print("skipping", i, "already labeled", clusters.loc[i, 'label'], flush=True)
                time.sleep(0.01)
                continue

        try:
            time.sleep(0.01)
            messages=[
                system_prompt, {"role":"user", "content": batch[0]} # TODO hardcoded batch size
            ]
            label = model.chat(messages)
            labels.append(label)
            # print("label:\n", label)
            # do some cleanup of the labels when the model doesn't follow instructions
            clean_label = label.replace("\n", " ")
            clean_label = clean_label.replace('"', '')
            clean_label = clean_label.replace("'", '')
            # clean_label = clean_label.replace("-", '')
            clean_label = ' '.join(clean_label.split())
            clean_label = " ".join(clean_label.split(" ")[0:5])
            clean_labels.append(clean_label)
            
            # print("clean_label:\n", clean_label)
            clusters.loc[i, 'label'] = clean_label
            clusters.loc[i, 'label_raw'] = label
            clusters.loc[i, 'labeled'] = True
            # length = len(clean_labels) - 1
            # clusters_df.loc[unlabled_row:unlabled_row+length, 'label'] = clean_labels
            # clusters_df.loc[unlabled_row:unlabled_row+length, 'label_raw'] = labels
            # clusters_df.loc[unlabled_row:unlabled_row+length, 'labeled'] = [True for i in range(0, len(labels))]
            clusters.to_parquet(os.path.join(cluster_dir, f"{label_id}.parquet"))
            # update 

        except Exception as e: 
            print(batch[0])
            print("ERROR:", e)
            print("exiting")
            exit(1)

    print("labels:", len(labels))
    # add lables to slides df

    # write the df to parquet
    with open(os.path.join(cluster_dir,f"{label_id}.json"), 'w') as f:
        json.dump({
            "id": label_id,
            "cluster_id": cluster_id,
            "model_id": model_id, 
            "text_column": text_column,
            "context": context,
            "system_prompt": system_prompt,
            "max_tokens": max_tokens,
        }, f, indent=2)
    f.close()
    print("done with", label_id)

if __name__ == "__main__":
    main()



class EmbedModelProvider:
    def __init__(self, name, params):
        self.name = name
        self.params = params

    def load_model(self):
        raise NotImplementedError("This method should be implemented by subclasses.")

    def embed(self, text):
        raise NotImplementedError("This method should be implemented by subclasses.")

class ChatModelProvider:
    def __init__(self, name, params):
        self.name = name
        self.params = params

    def load_model(self):
        raise NotImplementedError("This method should be implemented by subclasses.")

    def chat(self, messages):
        raise NotImplementedError("This method should be implemented by subclasses.")




# Usage: python umapper.py <dataset_id> <model> <neighbors> <min_dist>
# Example: python umapper.py dadabase-curated BAAI_bge-small-en-v1.5 50 0.075
import os
import re
import sys
import json
import umap
import h5py
import pickle
import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from latentscope.util import get_data_dir

def main():
    parser = argparse.ArgumentParser(description='UMAP embeddings for a dataset')
    parser.add_argument('dataset_id', type=str, help='Dataset name (directory name in data/)')
    parser.add_argument('embedding_id', type=str, help='Name of embedding model to use')
    parser.add_argument('neighbors', type=int, nargs="?", help='Output file', default=25)
    parser.add_argument('min_dist', type=float, nargs="?", help='Output file', default=0.075)
    parser.add_argument('--init', type=str, help='Initialize with UMAP', default=None)
    parser.add_argument('--align', type=str, help='Align UMAP with multiple embeddings', default=None)
    parser.add_argument('--save', action='store_true', help='Save the UMAP model')

    # Parse arguments
    args = parser.parse_args()
    umapper(args.dataset_id, args.embedding_id, args.neighbors, args.min_dist, save=args.save, init=args.init, align=args.align)


# TODO move this into shared space
def calculate_point_size(num_points, min_size=10, max_size=30, base_num_points=100):
    """
    Calculate the size of points for a scatter plot based on the number of points.
    """
    # TODO fix this to actually calculate a log scale between min and max size
    if num_points <= base_num_points:
        return max_size
    else:
        return min(min_size + min_size * np.log(num_points / base_num_points), max_size)


def umapper(dataset_id, embedding_id, neighbors=25, min_dist=0.1, save=False, init=None, align=None):
    DATA_DIR = get_data_dir()
    # read in the embeddings
    print("loading embeddings")
    embedding_path = os.path.join(DATA_DIR, dataset_id, "embeddings", f"{embedding_id}.h5")
    with h5py.File(embedding_path, 'r') as f:
        dataset = f["embeddings"]
        embeddings = np.array(dataset)

    umap_dir = os.path.join(DATA_DIR, dataset_id, "umaps")
    if not os.path.exists(umap_dir):
        os.makedirs(umap_dir)

    # determine the index of the last umap run by looking in the dataset directory
    # for files named umap-<number>.json
    umap_files = [f for f in os.listdir(umap_dir) if re.match(r"umap-\d+\.json", f)]
    if len(umap_files) > 0:
        last_umap = sorted(umap_files)[-1]
        last_umap_number = int(last_umap.split("-")[1].split(".")[0])
        next_umap_number = last_umap_number + 1
    else:
        next_umap_number = 1

    # make the umap name from the number, zero padded to 3 digits
    umap_id = f"umap-{next_umap_number:03d}"
    print("RUNNING:", umap_id)

    def process_umap_embeddings(umap_id, umap_embeddings, emb_id, align_id=None):
        min_values = np.min(umap_embeddings, axis=0)
        max_values = np.max(umap_embeddings, axis=0)

        # Scale the embeddings to the range [0, 1]
        umap_embeddings = (umap_embeddings - min_values) / (max_values - min_values)

        # Scale the embeddings to the range [-1, 1]
        umap_embeddings = 2 * umap_embeddings - 1

        print("writing normalized umap", umap_id)
        # save umap embeddings to a parquet file with columns x,y
        df = pd.DataFrame(umap_embeddings, columns=['x', 'y'])
        output_file = os.path.join(umap_dir, f"{umap_id}.parquet")
        df.to_parquet(output_file)
        print("wrote", output_file)

        # generate a scatterplot of the umap embeddings and save it to a file
        fig, ax = plt.subplots(figsize=(14.22, 14.22))  # 1024px by 1024px at 72 dpi
        point_size = calculate_point_size(umap_embeddings.shape[0])
        print("POINT SIZE", point_size, "for", umap_embeddings.shape[0], "points")
        plt.scatter(umap_embeddings[:, 0], umap_embeddings[:, 1], s=point_size, alpha=0.5)
        plt.axis('off')  # remove axis
        plt.gca().set_position([0, 0, 1, 1])  # remove margins
        plt.savefig(os.path.join(umap_dir, f"{umap_id}.png"))

        # save a json file with the umap parameters
        with open(os.path.join(umap_dir, f'{umap_id}.json'), 'w') as f:
            meta = {
                "id": umap_id, 
                "embedding_id": emb_id,
                "neighbors": neighbors, 
                "min_dist": min_dist,
            }
            if init is not None and init != "":
                meta["init"] = init,
            if align is not None and align != "":
                meta["align"] = f"{embedding_id},{align}"
                meta["align_id"] = align_id
            json.dump(meta, f, indent=2)
        f.close()


    if align is not None and align != "":
        print("aligned umap", align)
        # split the align string into umap names
        embs = align.split(",")
        # load each embedding from its h5 file
        a_embedding_ids = [embedding_id]
        a_embeddings = [embeddings]
        for emb in embs:
            print("loading", emb)
            emb_path = os.path.join(DATA_DIR, dataset_id, "embeddings", f"{emb}.h5")
            with h5py.File(emb_path, 'r') as f:
                dataset = f["embeddings"]
                a_emb = np.array(dataset)
            print("loaded", emb, "shape", a_emb.shape)
            a_embeddings.append(a_emb)
            a_embedding_ids.append(emb)
        
        reducer = umap.AlignedUMAP(
            n_neighbors=neighbors,
            min_dist=min_dist,
            metric='cosine',
            random_state=42,
            n_components=2,
            verbose=True,
        )
        print("a_embeddings", len(a_embeddings), len(a_embeddings[0]))
        relations = [{j: j for j in range(len(a_embeddings[i]))} for i in range(len(a_embeddings)-1)]
        print("relations", len(relations))
        aligned = reducer.fit_transform(a_embeddings, relations=relations)
        print("ALIGNED", aligned)
        for i,emb in enumerate(a_embedding_ids):
            print("processing", emb, "umap", next_umap_number+i)
            process_umap_embeddings(f"umap-{next_umap_number+i:03d}", aligned[i], emb, umap_id)

        print("done with aligned umap")
        return 

    if init is not None and init != "":
        print("loading umap", init)
        initial_df = pd.read_parquet(os.path.join(umap_dir, f"{init}.parquet"))
        initial = initial_df.to_numpy()
        print("initial shape", initial.shape)
        reducer = umap.UMAP(
            init=initial,
            n_neighbors=neighbors,
            min_dist=min_dist,
            metric='cosine',
            random_state=42,
            n_components=2,
            verbose=True,
        )
    else:
        reducer = umap.UMAP(
            n_neighbors=neighbors,
            min_dist=min_dist,
            metric='cosine',
            random_state=42,
            n_components=2,
            verbose=True,
        )
    print("reducing", embeddings.shape[0], "embeddings to 2 dimensions")
    umap_embeddings = reducer.fit_transform(embeddings)
    process_umap_embeddings(umap_id, umap_embeddings, embedding_id)

    if save:
        # save a pickle of the umap
        with open(os.path.join(umap_dir, f'{umap_id}.pkl'), 'wb') as f:
            pickle.dump(reducer, f)
    print("done with", umap_id)


if __name__ == "__main__":
    main()



import os
import time
import json
import uuid
import subprocess
import threading
from datetime import datetime
from flask import Blueprint, jsonify, request

# Create a Blueprint
jobs_bp = Blueprint('jobs_bp', __name__)
jobs_write_bp = Blueprint('jobs_write_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

TIMEOUT = 60 * 5 # 5 minute timeout TODO: make this a config option

PROCESSES = {}

def run_job(dataset, job_id, command):
    job_dir = os.path.join(DATA_DIR, dataset, "jobs")
    if not os.path.exists(job_dir):
      os.makedirs(job_dir)

    progress_file = os.path.join(job_dir, f"{job_id}.json")
    print("command", command)
    job_name = command.split(" ")[0]
    if "ls-" in job_name:
        job_name = job_name.replace("ls-", "")
    job = {
        "id": job_id,
        "dataset": dataset,
        "job_name": job_name,
        "command": command, 
        "status": "running", 
        "last_update": str(datetime.now()), 
        "progress": [], 
        "times": []
    }

    with open(progress_file, 'w') as f:
        json.dump(job, f)

    # TODO: need to watch for exploits in command if using shell=True for security reasons
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, text=True, shell=True)
    PROCESSES[job_id] = process

    last_output_time = time.time()  # Initialize with the current time

    while True:
        output = process.stdout.readline()
        current_time = time.time()  # Update current time on each iteration
        print(current_time, current_time - last_output_time, TIMEOUT)
        print("output", output)

        if output == '' and process.poll() is not None:
            break
        if output:
            print(output.strip())
            if("RUNNING:" in output):
                run_id = output.strip().split("RUNNING: ")[1]
                print("found the id", run_id)
                job["run_id"] = run_id
            job["progress"].append(output.strip())
            job["times"].append(str(datetime.now()))
            job["last_update"] = str(datetime.now())
            with open(progress_file, 'w') as f:
                json.dump(job, f)
            last_output_time = current_time
        elif current_time - last_output_time > TIMEOUT:
            print(f"Timeout: No output for more than {TIMEOUT} seconds.")
            print("OUTPUT", output)
            job["progress"].append(output.strip())
            job["progress"].append(f"Timeout: No output for more than {TIMEOUT} seconds.")
            job["status"] = "error"
            break  # Break the loop

    if process.returncode != 0:
        job["status"] = "error"
    else:
        job["status"] = "completed"
    # job["status"] = "completed"

    with open(progress_file, 'w') as f:
        json.dump(job, f)

@jobs_bp.route('/job')
def get_job():
    dataset = request.args.get('dataset')
    job_id = request.args.get('job_id')
    progress_file = os.path.join(DATA_DIR, dataset, "jobs", f"{job_id}.json")
    if os.path.exists(progress_file):
        try:
            with open(progress_file, 'r') as f:
                job = json.load(f)
        except:
            time.sleep(0.1)
            with open(progress_file, 'r') as f:
                job = json.load(f)
        return jsonify(job)
    else:
        return jsonify({'status': 'not found'}), 404

@jobs_bp.route('/all')
def get_jobs():
    dataset = request.args.get('dataset')
    job_dir = os.path.join(DATA_DIR, dataset, "jobs")
    jobs = []
    if os.path.exists(job_dir):
        for file in os.listdir(job_dir):
            if file.endswith(".json"):
                with open(os.path.join(job_dir, file), 'r') as f:
                    job = json.load(f)
                jobs.append(job)
    return jsonify(jobs)

@jobs_write_bp.route('/ingest', methods=['POST'])
def run_ingest():
    dataset = request.form.get('dataset')
    file = request.files.get('file')
    dataset_dir = os.path.join(DATA_DIR, dataset)
    if not os.path.exists(dataset_dir):
        os.makedirs(dataset_dir)
    file_path = os.path.join(dataset_dir, file.filename)
    file.save(file_path)

    job_id = str(uuid.uuid4())
    command = f'ls-ingest {dataset} --path={file_path}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})


@jobs_write_bp.route('/embed')
def run_embed():
    dataset = request.args.get('dataset')
    text_column = request.args.get('text_column')
    model_id = request.args.get('model_id') # model id
    prefix = request.args.get('prefix')
    dimensions = request.args.get('dimensions')

    job_id = str(uuid.uuid4())
    command = f'ls-embed {dataset} {text_column} {model_id} --prefix="{prefix}"'
    if dimensions is not None:
        command += f" --dimensions={dimensions}"
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/rerun')
def rerun_job():
    dataset = request.args.get('dataset')
    job_id = request.args.get('job_id')
    # read the job file to get the command
    progress_file = os.path.join(DATA_DIR, dataset, "jobs", f"{job_id}.json")
    with open(progress_file, 'r') as f:
        job = json.load(f)
    command = job.get('command')
    command += f' --rerun {job.get("run_id")}'
    new_job_id = str(uuid.uuid4())
    print("new job id", new_job_id)
    threading.Thread(target=run_job, args=(dataset, new_job_id, command)).start()
    return jsonify({"job_id": new_job_id})

@jobs_write_bp.route('/kill')
def kill_job():
    dataset = request.args.get('dataset')
    job_id = request.args.get('job_id')
    # load the job file
    progress_file = os.path.join(DATA_DIR, dataset, "jobs", f"{job_id}.json")
    job = json.load(open(progress_file, 'r'))
    if job_id in PROCESSES:
        PROCESSES[job_id].kill()
        job["status"] = "dead"
        job["cause_of_death"] = "killed"
        with open(progress_file, 'w') as f:
            json.dump(job, f)
        return jsonify(job)
    else:
        job["status"] = "dead"
        job["cause_of_death"] = "process not found, presumed dead"
        with open(progress_file, 'w') as f:
            json.dump(job, f)
        return jsonify(job)

@jobs_write_bp.route('/delete/embedding')
def delete_embedding():
    dataset = request.args.get('dataset')
    embedding_id = request.args.get('embedding_id')

    # Get a list of all the umaps that have embedding_id in their .json so we can delete them too
    umap_dir = os.path.join(DATA_DIR, dataset, 'umaps')
    umaps_to_delete = []
    for file in os.listdir(umap_dir):
        if file.endswith(".json"):
            with open(os.path.join(umap_dir, file), 'r') as f:
                umap_data = json.load(f)
            if umap_data.get('embedding_id') == embedding_id:
                umaps_to_delete.append(file.replace('.json', ''))
    

    job_id = str(uuid.uuid4())
    command = f'rm -rf {os.path.join(DATA_DIR, dataset, "embeddings", f"{embedding_id}*")}'
    for umap in umaps_to_delete:
        delete_umap(dataset, umap)
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/umap')
def run_umap():
    dataset = request.args.get('dataset')
    embedding_id = request.args.get('embedding_id')
    neighbors = request.args.get('neighbors')
    min_dist = request.args.get('min_dist')
    init = request.args.get('init')
    print("run umap", dataset, embedding_id, neighbors, min_dist, init)

    job_id = str(uuid.uuid4())
    command = f'ls-umap {dataset} {embedding_id} {neighbors} {min_dist} --init={init}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/delete/umap')
def delete_umap_request():
    dataset = request.args.get('dataset')
    umap_id = request.args.get('umap_id')
    return delete_umap(dataset, umap_id)

def delete_umap(dataset, umap_id):
    # Get a list of all the clusters that have umap_name in their .json so we can delete them too
    cluster_dir = os.path.join(DATA_DIR, dataset, 'clusters')
    clusters_to_delete = []
    for file in os.listdir(cluster_dir):
        if file.endswith(".json"):
            with open(os.path.join(cluster_dir, file), 'r') as f:
                cluster_data = json.load(f)
            if cluster_data.get('umap_id') == umap_id:
                clusters_to_delete.append(file.replace('.json', ''))
    

    job_id = str(uuid.uuid4())
    command = f'rm -rf {os.path.join(DATA_DIR, dataset, "umaps", f"{umap_id}*")}'
    # Create the rm -rf commands from the clusters_to_delete list
    for cluster in clusters_to_delete:
        command += f'; rm {os.path.join(DATA_DIR, dataset, "clusters", f"{cluster}*")}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/cluster')
def run_cluster():
    dataset = request.args.get('dataset')
    umap_id = request.args.get('umap_id')
    samples = request.args.get('samples')
    min_samples = request.args.get('min_samples')
    print("run cluster", dataset, umap_id, samples, min_samples)

    job_id = str(uuid.uuid4())
    command = f'ls-cluster {dataset} {umap_id} {samples} {min_samples}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/delete/cluster')
def delete_cluster():
    dataset = request.args.get('dataset')
    cluster_id = request.args.get('cluster_id')
    job_id = str(uuid.uuid4())
    command = f'rm -rf {os.path.join(DATA_DIR, dataset, "clusters", f"{cluster_id}*")}'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})

@jobs_write_bp.route('/cluster_label')
def run_cluster_label():
    dataset = request.args.get('dataset')
    chat_id = request.args.get('chat_id')
    text_column = request.args.get('text_column')
    cluster_id = request.args.get('cluster_id')
    context = request.args.get('context')
    print("run cluster label", dataset, chat_id, text_column, cluster_id)
    print("context", context)

    job_id = str(uuid.uuid4())
    command = f'ls-label {dataset} {text_column} {cluster_id} {chat_id} "{context}"'
    threading.Thread(target=run_job, args=(dataset, job_id, command)).start()
    return jsonify({"job_id": job_id})



import argparse
from latentscope.util import update_data_dir

def start():
  parser = argparse.ArgumentParser(description='Start the Latent Scope API')
  parser.add_argument('data_dir', type=str, nargs='?', default=None, help='Path to the directory where data is stored')
  parser.add_argument('--host', type=str, default="0.0.0.0", help='Host to serve the server on')
  parser.add_argument('--port', type=int, default=5001, help='Port to run the server on')
  parser.add_argument('--debug', action='store_true', help='Run server in debug mode')

  args = parser.parse_args()
  # This sets the environment variable expected
  update_data_dir(args.data_dir)
  from .app import serve
  serve(args.host, args.port, args.debug)

def serve(host="0.0.0.0", port=5001, debug=False):
  from .app import serve
  serve(host, port, debug)


import os
import sys
import numpy as np
from flask import Blueprint, jsonify, request

# Create a Blueprint
tags_bp = Blueprint('tags_bp', __name__)
tags_write_bp = Blueprint('tags_write_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

# ===========================================================
# Tags
# ===========================================================

# cache of list of indices per tag per dataset
tagsets = {}

"""
Return the tagsets for a given dataset
This is a JSON object with the tag name as the key and an array of indices as the value
"""
@tags_bp.route("/", methods=['GET'])
def tags():
    dataset = request.args.get('dataset')
    tagdir = os.path.join(DATA_DIR, dataset, "tags")
    if not os.path.exists(tagdir):
        os.makedirs(tagdir)
    if dataset not in tagsets:
        tagsets[dataset] = {}
    # search the dataset directory for all files ending in .indices
    for f in os.listdir(tagdir):
        if f.endswith(".indices"):
            tag = f.split(".")[0]
            indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), dtype=int).tolist()
            if type(indices) == int:
                indices = [indices]
            tagsets[dataset][tag] = indices

    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])

"""
Create a new tag for a given dataset
"""
@tags_write_bp.route("/new", methods=['GET'])
def new_tag():
    dataset = request.args.get('dataset')
    tag = request.args.get('tag')
    if dataset not in tagsets:
        tagsets[dataset] = {}
    # search the dataset directory for all files ending in .indices
    tags = []
    for f in os.listdir(os.path.join(DATA_DIR, dataset)):
        if f.endswith(".indices"):
            dtag = f.split(".")[0]
            indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", dtag + ".indices"), dtype=int).tolist()
            if type(indices) == int:
                indices = [indices]
            tagsets[dataset][dtag] = indices

    if tag not in tagsets[dataset]:
        tagsets[dataset][tag] = []
        # create an empty file
        filename = os.path.join(DATA_DIR, dataset, "tags", tag + ".indices")
        with open(filename, 'w') as f:
            f.write("")
            f.close()


    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])

"""
Add a data index to a tag
"""
@tags_write_bp.route("/add", methods=['GET'])
def add_tag():
    dataset = request.args.get('dataset')
    tag = request.args.get('tag')
    index = request.args.get('index')
    if dataset not in tagsets:
        ts = tagsets[dataset] = {}
    else:
        ts = tagsets[dataset]
    if tag not in ts:
        # read a tag file, which is just a csv with a single column into an array of integers
        indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), dtype=int).tolist()
        if type(indices) == int:
            indices = [indices]
        ts[tag] = indices
    else:
        indices = ts[tag]

    if not indices:
        indices = []
    if index not in indices:
        indices.append(int(index))
        ts[tag] = indices
        # save the indices to a file
        np.savetxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), indices, fmt='%d')
    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])

"""
Remove a data index from a tag
"""
@tags_write_bp.route("/remove", methods=['GET'])
def remove_tag():
    dataset = request.args.get('dataset')
    tag = request.args.get('tag')
    index = request.args.get('index')
    if dataset not in tagsets:
        tagsets[dataset] = {}
    else:
        ts = tagsets[dataset]
    if tag not in ts:
        # read a tag file, which is just a csv with a single column into an array of integers
        indices = np.loadtxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), dtype=int).tolist()
        if type(indices) == int:
            indices = [indices]
        ts[tag] = indices
    else:
        indices = ts[tag]
    if index in indices:
        indices = indices.remove(int(index))
        ts[tag] = indices
        # save the indices to a file
        np.savetxt(os.path.join(DATA_DIR, dataset, "tags", tag + ".indices"), indices, fmt='%d')
    # return an object with the tags for a given dataset
    return jsonify(tagsets[dataset])



import os
import re
import json
import fnmatch
import pandas as pd
from flask import Blueprint, jsonify, request

# Create a Blueprint
datasets_bp = Blueprint('datasets_bp', __name__)
datasets_write_bp = Blueprint('datasets_write_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

"""
Get the essential metadata for all available datasets.
Essential metadata is stored in meta.json
"""
@datasets_bp.route('/', methods=['GET'])
def get_datasets():
    datasets = []

    for dir in os.listdir(DATA_DIR):
        file_path = os.path.join(DATA_DIR, dir, 'meta.json')
        if os.path.isfile(file_path):
            with open(file_path, 'r', encoding='utf-8') as file:
                jsonData = json.load(file)
                jsonData['id'] = dir
                datasets.append(jsonData)

    datasets.sort(key=lambda x: x.get('length'))
    return jsonify(datasets)

"""
Get all metadata files from the given a directory.
"""
def scan_for_json_files(directory_path, match_pattern=r".*\.json$"):
    try:
        # files = os.listdir(directory_path)
        files = sorted(os.listdir(directory_path), key=lambda x: os.path.getmtime(os.path.join(directory_path, x)), reverse=True)
    except OSError as err:
        print('Unable to scan directory:', err)
        return jsonify({"error": "Unable to scan directory"}), 500

    json_files = [file for file in files if re.match(match_pattern, file)]
    # print("files", files)
    # print("json", json_files)

    json_contents = []
    for file in json_files:
        try:
            with open(os.path.join(directory_path, file), 'r', encoding='utf-8') as json_file:
                json_contents.append(json.load(json_file))
        except json.JSONDecodeError as err:
            print('Error parsing JSON string:', err)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/meta', methods=['GET'])
def get_dataset_meta(dataset):
    file_path = os.path.join(DATA_DIR, dataset, "meta.json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

@datasets_write_bp.route('/<dataset>/meta/update', methods=['GET'])
def update_dataset_meta(dataset):
    key = request.args.get('key')
    value = request.args.get('value')
    file_path = os.path.join(DATA_DIR, dataset, "meta.json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    json_contents[key] = value
    # write the file back out
    with open(file_path, 'w', encoding='utf-8') as json_file:
        json.dump(json_contents, json_file)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/embeddings', methods=['GET'])
def get_dataset_embeddings(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "embeddings")
    # directory_path = os.path.join(DATA_DIR, dataset, "umaps")
    return scan_for_json_files(directory_path)

@datasets_bp.route('/<dataset>/umaps', methods=['GET'])
def get_dataset_umaps(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "umaps")
    return scan_for_json_files(directory_path)

@datasets_bp.route('/<dataset>/umaps/<umap>', methods=['GET'])
def get_dataset_umap(dataset, umap):
    file_path = os.path.join(DATA_DIR, dataset, "umaps", umap + ".json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

@datasets_bp.route('/<dataset>/umaps/<umap>/points', methods=['GET'])
def get_dataset_umap_points(dataset, umap):
    file_path = os.path.join(DATA_DIR, dataset, "umaps", umap + ".parquet")
    df = pd.read_parquet(file_path)
    return df.to_json(orient="records")

@datasets_bp.route('/<dataset>/clusters', methods=['GET'])
def get_dataset_clusters(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "clusters")
    return scan_for_json_files(directory_path, match_pattern=r"cluster-\d+\.json")

@datasets_bp.route('/<dataset>/clusters/<cluster>', methods=['GET'])
def get_dataset_cluster(dataset, cluster):
    file_path = os.path.join(DATA_DIR, dataset, "clusters", cluster + ".json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

# @datasets_bp.route('/<dataset>/clusters/<cluster>/labels', methods=['GET'])
# def get_dataset_cluster_labels_default(dataset, cluster):
#     file_name = cluster + "-labels.parquet"
#     file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
#     df = pd.read_parquet(file_path)
#     return df.to_json(orient="records")

@datasets_bp.route('/<dataset>/clusters/<cluster>/indices', methods=['GET'])
def get_dataset_cluster_indices(dataset, cluster):
    file_name = cluster + ".parquet"
    file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
    df = pd.read_parquet(file_path)
    return df.to_json(orient="records")

@datasets_bp.route('/<dataset>/clusters/<cluster>/labels/<id>', methods=['GET'])
def get_dataset_cluster_labels(dataset, cluster, id):
    # if model == "default":
    #     return get_dataset_cluster_labels_default(dataset, cluster)
    file_name = cluster + "-labels-" + id + ".parquet"
    file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
    df = pd.read_parquet(file_path)
    df.reset_index(inplace=True)
    return df.to_json(orient="records")

@datasets_write_bp.route('/<dataset>/clusters/<cluster>/labels/<id>/label/<index>', methods=['GET'])
def overwrite_dataset_cluster_label(dataset, cluster, id, index):
    index = int(index)
    new_label = request.args.get('label')
    print("write label", index, new_label)
    if new_label is None:
        return jsonify({"error": "Missing 'label' in request data"}), 400

    file_name = cluster + "-labels-" + id + ".parquet"
    file_path = os.path.join(DATA_DIR, dataset, "clusters", file_name)
    try:
        df = pd.read_parquet(file_path)
    except FileNotFoundError:
        return jsonify({"error": "File not found"}), 404

    if index >= len(df):
        return jsonify({"error": "Index out of range"}), 400

    df.at[index, 'label'] = new_label
    df.to_parquet(file_path)

    return jsonify({"success": True, "message": "Label updated successfully"})


@datasets_bp.route('/<dataset>/clusters/<cluster>/labels_available', methods=['GET'])
def get_dataset_cluster_labels_available(dataset, cluster):
    directory_path = os.path.join(DATA_DIR, dataset, "clusters")
    return scan_for_json_files(directory_path, match_pattern=rf"{cluster}-labels-.*\.json")
    # try:
    #     files = sorted(os.listdir(directory_path), key=lambda x: os.path.getmtime(os.path.join(directory_path, x)), reverse=True)
    # except OSError as err:
    #     print('Unable to scan directory:', err)
    #     return jsonify({"error": "Unable to scan directory"}), 500

    # pattern = re.compile(r'^' + cluster + '-labels-(.*).parquet$')
    # model_names = [pattern.match(file).group(1) for file in files if pattern.match(file)]
    # return jsonify(model_names)


def get_next_scopes_number(dataset):
    # figure out the latest scope number
    scopes_files = [f for f in os.listdir(os.path.join(DATA_DIR,dataset,"scopes")) if re.match(r"scopes-\d+\.json", f)]
    if len(scopes_files) > 0:
        last_scopes = sorted(scopes_files)[-1]
        last_scopes_number = int(last_scopes.split("-")[1].split(".")[0])
        next_scopes_number = last_scopes_number + 1
    else:
        next_scopes_number = 1
    return next_scopes_number

@datasets_bp.route('/<dataset>/scopes', methods=['GET'])
def get_dataset_scopes(dataset):
    directory_path = os.path.join(DATA_DIR, dataset, "scopes")
    print("dataset", dataset, directory_path)
    return scan_for_json_files(directory_path)

@datasets_bp.route('/<dataset>/scopes/<scope>', methods=['GET'])
def get_dataset_scope(dataset, scope):
    directory_path = os.path.join(DATA_DIR, dataset, "scopes")
    file_path = os.path.join(directory_path, scope + ".json")
    with open(file_path, 'r', encoding='utf-8') as json_file:
        json_contents = json.load(json_file)
    return jsonify(json_contents)

@datasets_write_bp.route('/<dataset>/scopes/save', methods=['POST'])
def save_dataset_scope(dataset):
    if not request.json:
        return jsonify({"error": "Invalid data format, JSON expected"}), 400
    id = request.json.get('id')
    embedding_id = request.json.get('embedding_id')
    umap_id = request.json.get('umap_id')
    cluster_id = request.json.get('cluster_id')
    cluster_labels_id = request.json.get('cluster_labels_id')
    label = request.json.get('label')
    description = request.json.get('description')
    scope = {
        "embedding_id": embedding_id,
        "umap_id": umap_id,
        "cluster_id": cluster_id,
        "cluster_labels_id": cluster_labels_id,
        "label": label,
        "description": description
    }
    if not id:
        next_scopes_number = get_next_scopes_number(dataset)
        # make the umap name from the number, zero padded to 3 digits
        id = f"scopes-{next_scopes_number:03d}"
    scope["id"] = id
    file_path = os.path.join(DATA_DIR, dataset, "scopes", id + ".json")
    with open(file_path, 'w') as f:
        json.dump(scope, f, indent=2)
    return jsonify(scope)



# Usage: python cluster.py <dataset_name> <umap_name> <samples> 
# Example: python cluster.py dadabase-curated umap1d-001 50
import os
import re
import sys
import json
import hdbscan
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.spatial.distance import cdist


def clusterer(dataset_name, umap_name, samples, min_samples):

    # determine the index of the last cluster run by looking in the dataset directory
    # for files named umap1d-<number>.json
    cluster_files = [f for f in os.listdir(f"../data/{dataset_name}") if re.match(r"cluster-\d+\.json", f)]
    print("cluster files", sorted(cluster_files))
    if len(cluster_files) > 0:
        last_cluster = sorted(cluster_files)[-1]
        last_cluster_number = int(last_cluster.split("-")[1].split(".")[0])
        print("lastcluster", last_cluster, last_cluster_number)
        next_cluster_number = last_cluster_number + 1
    else:
        next_cluster_number = 1

    # make the umap name from the number, zero padded to 3 digits
    cluster_name = f"cluster-{next_cluster_number:03d}"

    # save a json file with the umap parameters
    with open(f'../data/{dataset_name}/clusters/{cluster_name}.json', 'w') as f:
        json.dump({"umap_name": umap_name, "samples": samples, "min_samples": min_samples}, f)

    umap_embeddings_df = pd.read_parquet(f"../data/{dataset_name}/umaps/{umap_name}.parquet")
    umap_embeddings = umap_embeddings_df.to_numpy()

    clusterer = hdbscan.HDBSCAN(min_cluster_size=samples, min_samples=min_samples, metric='euclidean')
    clusterer.fit(umap_embeddings)

    # Get the cluster labels
    cluster_labels = clusterer.labels_
    # copy cluster labels to another array
    raw_cluster_labels = cluster_labels.copy()

    # Determine points with no assigned cluster
    unique_labels = np.unique(cluster_labels)
    non_noise_labels = unique_labels[unique_labels != -1]
    centroids = [umap_embeddings[cluster_labels == label].mean(axis=0) for label in non_noise_labels]

    # Assign noise points to the closest cluster centroid
    noise_points = umap_embeddings[cluster_labels == -1]
    if(non_noise_labels.shape[0] > 0):
      closest_centroid_indices = np.argmin(cdist(noise_points, centroids), axis=1)

      # Update cluster_labels with the new assignments for noise points
      noise_indices = np.where(cluster_labels == -1)[0]
      new_assignments = [non_noise_labels[index] for index in closest_centroid_indices]
      cluster_labels[noise_indices] = new_assignments

    with open(f'../data/{dataset_name}/clusters/{cluster_name}.json', 'w') as f:
        json.dump({
            "umap_name": umap_name, 
            "samples": samples, 
            "min_samples": min_samples,
            "n_clusters": len(non_noise_labels),
            "n_noise": len(noise_points)
        }, f)
    print("n_clusters:", len(non_noise_labels))
    print("noise points assigned to clusters:", len(noise_points))

    # save umap embeddings to a parquet file with columns x,y
    df = pd.DataFrame({"cluster": cluster_labels, "raw_cluster": raw_cluster_labels})
    output_file = f"../data/{dataset_name}/clusters/{cluster_name}.parquet"
    df.to_parquet(output_file)
    print(df.head())

    # generate a scatterplot of the umap embeddings and save it to a file
    fig, ax = plt.subplots(figsize=(6, 6))
    y_values = np.full(len(umap_embeddings), 0.5)
    plt.scatter(umap_embeddings[:, 0], y_values, s=1, alpha=0.5, c=cluster_labels, cmap='Spectral')
    plt.axis('off')  # remove axis
    plt.gca().set_position([0, 0, 1, 1])  # remove margins
    plt.savefig(f"../data/{dataset_name}/clusters/{cluster_name}.png")

    print("wrote", output_file)

if __name__ == "__main__":
    dataset_name = sys.argv[1]
    umap_name = sys.argv[2]
    samples = int(sys.argv[3])
    min_samples = 5
    clusterer(dataset_name, umap_name, samples, min_samples)



import os
import time
import voyageai 
import tokenizers
from tokenizers import Tokenizer
from .base import EmbedModelProvider


from latentscope.util import get_key

class VoyageAIEmbedProvider(EmbedModelProvider):
    def load_model(self):
        api_key = get_key("VOYAGE_API_KEY")
        if api_key is None:
            print("ERROR: No API key found for Voyage")
            print("Missing 'VOYAGE_API_KEY' variable in:", f"{os.getcwd()}/.env")
        self.client = voyageai.Client(api_key)
        # The voyage client provides a tokenizer that only encodes https://docs.voyageai.com/tokenization/
        # It also says that it uses the same tokenizer as Llama 2
        self.encoder = Tokenizer.from_pretrained("TheBloke/Llama-2-70B-fp16")

    def embed(self, inputs):
        time.sleep(0.1) # TODO proper rate limiting
        # We truncate the input ourselves, even though the API supports truncation its still possible to send too big a batch
        enc = self.encoder
        max_tokens = self.params["max_tokens"]
        inputs = [enc.decode(enc.encode(b).ids[:max_tokens]) if len(enc.encode(b)) > max_tokens else b for b in inputs]
        response = self.client.embed(texts=inputs, model=self.name, truncation=self.params["truncation"])
        embeddings = response.embeddings
        return embeddings


import os
import time
import tiktoken
import together
from .base import EmbedModelProvider

from latentscope.util import get_key

class TogetherAIEmbedProvider(EmbedModelProvider):
    def load_model(self):
        api_key = get_key("TOGETHER_API_KEY")
        if api_key is None:
            print("ERROR: No API key found for Together")
            print("Missing 'TOGETHER_API_KEY' variable in:", f"{os.getcwd()}/.env")
        together.api_key = api_key
        self.client = together.Together()
        self.encoder = tiktoken.encoding_for_model("text-embedding-ada-002")

    def embed(self, inputs):
        time.sleep(0.2) # TODO proper rate limiting
        enc = self.encoder
        max_tokens = self.params["max_tokens"]
        inputs = [b.replace("\n", " ") for b in inputs]
        inputs = [enc.decode(enc.encode(b)[:max_tokens]) if len(enc.encode(b)) > max_tokens else b for b in inputs]
        response = self.client.embeddings.create(
            input=inputs,
            model=self.name
        )
        embeddings = [response.data[i].embedding for i in range(len(inputs))]
        return embeddings


# Usage: python umapper.py <dataset_name> <neighbors> <min_dist>
# Example: python umapper.py dadabase-curated 50 0.075
# TODO: update this to match latest improvements in umapper.py
import os
import re
import sys
import json
import umap
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def umapper(dataset_name, neighbors=25, min_dist=0.075):
    # read in the embeddings
    embeddings = np.load(f'../data/{dataset_name}/embeddings.npy')

    # determine the index of the last umap run by looking in the dataset directory
    # for files named umap-<number>.json
    umap_files = [f for f in os.listdir(f"../data/{dataset_name}/umaps") if re.match(r"umap1d-\d+\.json", f)]
    if len(umap_files) > 0:
        last_umap = sorted(umap_files)[-1]
        last_umap_number = int(last_umap.split("-")[1].split(".")[0])
        next_umap_number = last_umap_number + 1
    else:
        next_umap_number = 1

    # make the umap name from the number, zero padded to 3 digits
    umap_name = f"umap1d-{next_umap_number:03d}"

    # save a json file with the umap parameters
    with open(f'../data/{dataset_name}/umaps/{umap_name}.json', 'w') as f:
        json.dump({"neighbors": neighbors, "min_dist": min_dist}, f)

    reducer = umap.UMAP(
        n_neighbors=neighbors,
        min_dist=min_dist,
        metric='cosine',
        random_state=42,
        n_components=1,
        verbose=True,
    )

    umap_embeddings = reducer.fit_transform(embeddings)

    min_values = np.min(umap_embeddings, axis=0)
    max_values = np.max(umap_embeddings, axis=0)

    # Scale the embeddings to the range [0, 1]
    umap_embeddings = (umap_embeddings - min_values) / (max_values - min_values)

    # Scale the embeddings to the range [-1, 1]
    # umap_embeddings = 2 * umap_embeddings - 1

    # save umap embeddings to a parquet file with columns x,y
    df = pd.DataFrame(umap_embeddings, columns=['x'])
    output_file = f"../data/{dataset_name}/umaps/{umap_name}.parquet"
    df.to_parquet(output_file)
    print("wrote", output_file)

    # generate a scatterplot of the umap embeddings and save it to a file
    
    fig, ax = plt.subplots(figsize=(6, 6))
    y_values = np.full(len(umap_embeddings), 0.5)
    plt.scatter(umap_embeddings[:, 0], y=y_values, s=1, alpha=0.5)
    plt.axis('off')  # remove axis
    plt.gca().set_position([0, 0, 1, 1])  # remove margins
    plt.savefig(f"../data/{dataset_name}/umaps/{umap_name}.png")
    


if __name__ == "__main__":
    dataset_name = sys.argv[1]
    neighbors = int(sys.argv[2])
    min_dist = float(sys.argv[3])
    umapper(dataset_name, neighbors=neighbors, min_dist=min_dist)



import os
import time
import tiktoken
from openai import OpenAI
from .base import EmbedModelProvider, ChatModelProvider

from latentscope.util import get_key

class OpenAIEmbedProvider(EmbedModelProvider):
    def load_model(self):
        api_key = get_key("OPENAI_API_KEY")
        if api_key is None:
            print("ERROR: No API key found for OpenAI")
            print("Missing 'OPENAI_API_KEY' variable in:", f"{os.getcwd()}/.env")
        self.client = OpenAI(api_key=api_key)
        # special case for the new embedding models
        if self.name in ["text-embedding-3-small", "text-embedding-3-large"]:
            self.encoder = tiktoken.get_encoding("cl100k_base")
        else:
            self.encoder = tiktoken.encoding_for_model(self.name)

    def embed(self, inputs, dimensions=None):
        time.sleep(0.01) # TODO proper rate limiting
        enc = self.encoder
        max_tokens = self.params["max_tokens"]
        inputs = [b.replace("\n", " ") for b in inputs]
        inputs = [enc.decode(enc.encode(b)[:max_tokens]) if len(enc.encode(b)) > max_tokens else b for b in inputs]
        if dimensions is not None and dimensions > 0:
            response = self.client.embeddings.create(
                input=inputs,
                model=self.name,
                dimensions=dimensions
            )
        else:
            response = self.client.embeddings.create(
                input=inputs,
                model=self.name
            )
        embeddings = [embedding.embedding for embedding in response.data]
        return embeddings

class OpenAIChatProvider(ChatModelProvider):
    def load_model(self):
        self.client = OpenAI(api_key=get_key("OPENAI_API_KEY"))
        self.encoder = tiktoken.encoding_for_model(self.name)

    def chat(self, messages):
        response = self.client.chat.completions.create(
            model=self.name,
            messages=messages
        )
        return response.choices[0].message.content


# Usage: python cluster.py <dataset_id> <umap_id> <samples> <min_samples>
# Example: python cluster.py dadabase-curated umap-001 50 5
import os
import re
import json
import hdbscan
import argparse
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.spatial import ConvexHull
from scipy.spatial.distance import cdist

from latentscope.util import get_data_dir

# TODO move this into shared space
def calculate_point_size(num_points, min_size=10, max_size=30, base_num_points=100):
    """
    Calculate the size of points for a scatter plot based on the number of points.
    """
    # TODO fix this to actually calculate a log scale between min and max size
    if num_points <= base_num_points:
        return max_size
    else:
        return min(min_size + min_size * np.log(num_points / base_num_points), max_size)


def main():
    parser = argparse.ArgumentParser(description='Cluster UMAP embeddings')
    parser.add_argument('dataset_id', type=str, help='ID of the dataset')
    parser.add_argument('umap_id', type=str, help='ID of the UMAP file')
    parser.add_argument('samples', type=int, help='Minimum cluster size')
    parser.add_argument('min_samples', type=int, help='Minimum samples for HDBSCAN')
    
    args = parser.parse_args()
    clusterer(args.dataset_id, args.umap_id, args.samples, args.min_samples)


def clusterer(dataset_id, umap_id, samples, min_samples):
    DATA_DIR = get_data_dir()
    cluster_dir = os.path.join(DATA_DIR, dataset_id, "clusters")
    # Check if clusters directory exists, if not, create it
    if not os.path.exists(cluster_dir):
        os.makedirs(cluster_dir)
    # determine the index of the last cluster run by looking in the dataset directory
    # for files named umap-<number>.json
    cluster_files = [f for f in os.listdir(cluster_dir) if re.match(r"cluster-\d+\.json", f)]
    if len(cluster_files) > 0:
        last_cluster = sorted(cluster_files)[-1]
        last_cluster_number = int(last_cluster.split("-")[1].split(".")[0])
        next_cluster_number = last_cluster_number + 1
    else:
        next_cluster_number = 1

    # make the umap name from the number, zero padded to 3 digits
    cluster_id = f"cluster-{next_cluster_number:03d}"
    print("RUNNING:", cluster_id)

    umap_embeddings_df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "umaps", f"{umap_id}.parquet"))
    umap_embeddings = umap_embeddings_df.to_numpy()

    clusterer = hdbscan.HDBSCAN(min_cluster_size=samples, min_samples=min_samples, metric='euclidean')
    clusterer.fit(umap_embeddings)

    # Get the cluster labels
    cluster_labels = clusterer.labels_
    # copy cluster labels to another array
    raw_cluster_labels = cluster_labels.copy()

    # Determine points with no assigned cluster
    unique_labels = np.unique(cluster_labels)
    non_noise_labels = unique_labels[unique_labels != -1]
    centroids = [umap_embeddings[cluster_labels == label].mean(axis=0) for label in non_noise_labels]

    # Assign noise points to the closest cluster centroid
    noise_points = umap_embeddings[cluster_labels == -1]
    if(non_noise_labels.shape[0] > 0):
      closest_centroid_indices = np.argmin(cdist(noise_points, centroids), axis=1)

      # Update cluster_labels with the new assignments for noise points
      noise_indices = np.where(cluster_labels == -1)[0]
      new_assignments = [non_noise_labels[index] for index in closest_centroid_indices]
      cluster_labels[noise_indices] = new_assignments

    
    print("n_clusters:", len(non_noise_labels))
    print("noise points assigned to clusters:", len(noise_points))

    # save umap embeddings to a parquet file with columns x,y
    df = pd.DataFrame({"cluster": cluster_labels, "raw_cluster": raw_cluster_labels})
    output_file = os.path.join(cluster_dir, f"{cluster_id}.parquet")
    df.to_parquet(output_file)
    print(df.head())
    print("wrote", output_file)

    # generate a scatterplot of the umap embeddings and save it to a file
    fig, ax = plt.subplots(figsize=(14.22, 14.22))  # 1024px by 1024px at 72 dpi
    point_size = calculate_point_size(umap_embeddings.shape[0])
    print("POINT SIZE", point_size, "for", umap_embeddings.shape[0], "points")
    plt.scatter(umap_embeddings[:, 0], umap_embeddings[:, 1], s=point_size, alpha=0.5, c=cluster_labels, cmap='Spectral')
    # plot a convex hull around each cluster
    hulls = []
    for label in non_noise_labels:
        indices = np.where(cluster_labels == label)[0]
        points = umap_embeddings[indices]
        # points = umap_embeddings[cluster_labels == label]
        hull = ConvexHull(points)
        hull_list = [indices[s] for s in hull.vertices.tolist()]
        hulls.append(hull_list)
        for simplex in hull.simplices:
            plt.plot(points[simplex, 0], points[simplex, 1], 'k-')

    plt.axis('off')  # remove axis
    plt.gca().set_position([0, 0, 1, 1])  # remove margins
    plt.savefig(os.path.join(cluster_dir, f"{cluster_id}.png"))

    with open(os.path.join(cluster_dir,f"{cluster_id}.json"), 'w') as f:
        json.dump({
            "id": cluster_id,
            "umap_id": umap_id, 
            "samples": samples, 
            "min_samples": min_samples,
            "n_clusters": len(non_noise_labels),
            "n_noise": len(noise_points)
        }, f, indent=2)
    f.close()

    # create the data structure for labeling clusters
    # get the indices of each item in a cluster
    cluster_indices = df.groupby('cluster').groups

    # iterate over the clusters and create a row for each in a new dataframe with a label, description and array of indicies
    slides_df = pd.DataFrame(columns=['label', 'description', 'indices'])
    for cluster, indices in cluster_indices.items():
        label = f"Cluster {cluster}"
        description = f"This is cluster {cluster} with {len(indices)} items."
        new_row = pd.DataFrame({'label': [label], 'description': [description], 'indices': [list(indices)], 'hull': [hulls[cluster]]})
        slides_df = pd.concat([slides_df, new_row], ignore_index=True)

    # write the df to parquet
    slides_df.to_parquet(os.path.join(cluster_dir, f"{cluster_id}-labels-default.parquet"))
    print("done with", cluster_id)

if __name__ == "__main__":
    main()


import os
import json
import pkg_resources
from .providers.transformers import TransformersEmbedProvider, TransformersChatProvider
from .providers.openai import OpenAIEmbedProvider, OpenAIChatProvider
from .providers.mistralai import MistralAIEmbedProvider, MistralAIChatProvider
from .providers.cohereai import CohereAIEmbedProvider
from .providers.togetherai import TogetherAIEmbedProvider
from .providers.voyageai import VoyageAIEmbedProvider

def get_embedding_model_list():
    """Returns a list of available embedding models."""
    embedding_path = pkg_resources.resource_filename('latentscope.models', 'embedding_models.json')
    with open(embedding_path, "r") as f:
        embed_model_list = json.load(f)
    return embed_model_list

def get_embedding_model_dict(id):
    embed_model_list = get_embedding_model_list()
    embed_model_dict = {model['id']: model for model in embed_model_list}
    model = embed_model_dict[id]
    if not model:
        raise ValueError(f"Model {id} not found")
    return model

def get_embedding_model(id):
    """Returns a ModelProvider instance for the given model id."""
    model = get_embedding_model_dict(id)
      
    if model['provider'] == "transformers":
        return TransformersEmbedProvider(model['name'], model['params'])
    if model['provider'] == "openai":
        return OpenAIEmbedProvider(model['name'], model['params'])
    if model['provider'] == "mistralai":
        return MistralAIEmbedProvider(model['name'], model['params'])
    if model['provider'] == "cohereai":
        return CohereAIEmbedProvider(model['name'], model['params'])
    if model['provider'] == "togetherai":
        return TogetherAIEmbedProvider(model['name'], model['params'])
    if model['provider'] == "voyageai":
        return VoyageAIEmbedProvider(model['name'], model['params'])


def get_chat_model_list():
    """Returns a list of available chat models."""
    chat_path = pkg_resources.resource_filename('latentscope.models', 'chat_models.json')
    with open(chat_path, "r") as f:
        chat_model_list = json.load(f)
    return chat_model_list

def get_chat_model_dict(id):
    chat_model_list = get_chat_model_list()
    chat_model_dict = {model['id']: model for model in chat_model_list}
    model = chat_model_dict[id]
    if not model:
        raise ValueError(f"Model {id} not found")
    return model

def get_chat_model(id):
    """Returns a ModelProvider instance for the given model id."""
    model = get_chat_model_dict(id)
    
    if model['provider'] == "transformers":
        return TransformersChatProvider(model['name'], model['params'])
    if model['provider'] == "openai":
        return OpenAIChatProvider(model['name'], model['params'])
    if model['provider'] == "mistralai":
        return MistralAIChatProvider(model['name'], model['params'])



import os
import time
from mistralai.client import MistralClient
from mistralai.models.chat_completion import ChatMessage
from transformers import AutoTokenizer
from .base import EmbedModelProvider,ChatModelProvider

from latentscope.util import get_key

# TODO verify these tokenizers somehow
# derived from:
  # https://docs.mistral.ai/platform/endpoints/
  # https://huggingface.co/docs/transformers/main/en/model_doc/mixtral
encoders = {
    "mistral-tiny": "mistralai/Mistral-7B-v0.1",
    "mistral-small": "mistralai/Mixtral-8x7B-v0.1",
    "mistral-medium": "mistralai/Mixtral-8x7B-v0.1", #just guessing
}

class MistralAIEmbedProvider(EmbedModelProvider):
    def load_model(self):
        api_key = get_key("MISTRAL_API_KEY")
        if api_key is None:
            print("ERROR: No API key found for Mistral")
            print("Missing 'MISTRAL_API_KEY' variable in:", f"{os.getcwd()}/.env")
        self.client = MistralClient(api_key=api_key)

    def embed(self, inputs):
        time.sleep(0.1) # TODO proper rate limiting
        response = self.client.embeddings(input=inputs, model=self.name)
        return [e.embedding for e in response.data]

class MistralAIChatProvider(ChatModelProvider):
    def load_model(self):
        api_key = get_key("MISTRAL_API_KEY")
        if api_key is None:
            print("ERROR: No API key found for Mistral")
            print("Missing 'MISTRAL_API_KEY' variable in:", f"{os.getcwd()}/.env")
        self.client = MistralClient(api_key=api_key)
        self.encoder = AutoTokenizer.from_pretrained(encoders[self.name])

    def chat(self, messages):
        instances = [ChatMessage(content=message["content"], role=message["role"]) for message in messages]
        response = self.client.chat(
            model=self.name,
            messages=instances
        )
        return response.choices[0].message.content


import os
import sys
from dotenv import load_dotenv, set_key

def get_data_dir():
    print("Loading environment variables from:", os.path.join(os.getcwd(), '.env'))
    load_dotenv()
    DATA_DIR = os.getenv('LATENT_SCOPE_DATA')
    if DATA_DIR is None:
        print("""LATENT_SCOPE_DATA environment variable not set. Please set it to the directory where you want to store your data.
e.g.: export LATENT_SCOPE_DATA=~/latentscope-data""")
        sys.exit(1)
    return DATA_DIR

def update_data_dir(directory, env_file=".env"):
    # Load existing .env file, or create one if it doesn't exist
    load_dotenv(env_file)
    if not directory or directory == "":
        directory = os.getenv('LATENT_SCOPE_DATA')
        if not directory:
            print("ERROR: Please specify a directory")
            return
        else:
            print("No directory specified, current directory is:", directory)
    if "~" in directory:
        directory = os.path.expanduser(directory)
    # Update the .env file with the new directory
    set_key(env_file, 'LATENT_SCOPE_DATA', directory)
    # Update the environment variable for the current process
    os.environ['LATENT_SCOPE_DATA'] = directory
    if not os.path.exists(directory):
        os.makedirs(directory)
    return directory

def get_key(key, env_file=".env"):
    print("get key", os.getcwd())
    load_dotenv(env_file)
    return os.getenv(key)

def set_openai_key(openai_key, env_file=".env"):
    # Load existing .env file, or create one if it doesn't exist
    load_dotenv(env_file)
    # Update the .env file with the new directory
    set_key(env_file, 'OPENAI_API_KEY', openai_key)
    # Update the environment variable for the current process
    os.environ['OPENAI_API_KEY'] = openai_key

def set_voyage_key(voyage_key, env_file=".env"):
    # Load existing .env file, or create one if it doesn't exist
    load_dotenv(env_file)
    # Update the .env file with the new directory
    set_key(env_file, 'VOYAGE_API_KEY', voyage_key)
    # Update the environment variable for the current process
    os.environ['VOYAGE_API_KEY'] = voyage_key

def set_together_key(together_key, env_file=".env"):
    # Load existing .env file, or create one if it doesn't exist
    load_dotenv(env_file)
    # Update the .env file with the new directory
    set_key(env_file, 'TOGETHER_API_KEY', together_key)
    # Update the environment variable for the current process
    os.environ['TOGETHER_API_KEY'] = together_key

def set_cohere_key(cohere_key, env_file=".env"):
    # Load existing .env file, or create one if it doesn't exist
    load_dotenv(env_file)
    # Update the .env file with the new directory
    set_key(env_file, 'COHERE_API_KEY', cohere_key)
    # Update the environment variable for the current process
    os.environ['COHERE_API_KEY'] = cohere_key

def set_mistral_key(mistral_key, env_file=".env"):
    # Load existing .env file, or create one if it doesn't exist
    load_dotenv(env_file)
    # Update the .env file with the new directory
    set_key(env_file, 'MISTRAL_API_KEY', mistral_key)
    # Update the environment variable for the current process
    os.environ['MISTRAL_API_KEY'] = mistral_key


# Usage: ls-embed <dataset_id> <text_column> <model_id>
import os
import re
import sys
import json
import time
import h5py
import argparse
import numpy as np
import pandas as pd

try:
    # Check if the runtime environment is a Jupyter notebook
    if 'ipykernel' in sys.modules and 'IPython' in sys.modules:
        from tqdm.notebook import tqdm
    else:
        from tqdm import tqdm
except ImportError as e:
    # Fallback to the standard console version if import fails
    from tqdm import tqdm

from latentscope.models import get_embedding_model, get_embedding_model_dict
from latentscope.util import get_data_dir

def chunked_iterable(iterable, size):
    """Yield successive chunks from an iterable."""
    for i in range(0, len(iterable), size):
        yield iterable[i:i + size]

def append_to_hdf5(file_path, new_data):
    dataset_name = "embeddings"
    with h5py.File(file_path, 'a') as f:
        if dataset_name in f:
            dataset = f[dataset_name]
            dataset.resize((dataset.shape[0] + new_data.shape[0],) + dataset.shape[1:])
            dataset[-new_data.shape[0]:] = new_data
        else:
            maxshape = (None,) + new_data.shape[1:]
            dataset = f.create_dataset(dataset_name, data=new_data, maxshape=maxshape, chunks=True)

def get_last_batch(file_path):
    try:
        with h5py.File(file_path, 'r') as f:
            dataset = f["embeddings"]
            return dataset.shape[0]
    except FileNotFoundError:
        return 0


def main():
    parser = argparse.ArgumentParser(description='Embed a dataset')
    parser.add_argument('dataset_id', type=str, help='Dataset id (directory name in data/)')
    parser.add_argument('text_column', type=str, help='Output file', default='text')
    parser.add_argument('model_id', type=str, help='ID of embedding model to use', default="transformers-BAAI___bge-small-en-v1.5")
    parser.add_argument('--prefix', type=str, help='Prefix to prepend to text before embedding', default="")
    parser.add_argument('--dimensions', type=int, help='Truncate embeddings to dimensions a la Matroyshka embeddings')
    parser.add_argument('--rerun', type=str, help='Rerun the given embedding from last completed batch')

    # Parse arguments
    args = parser.parse_args()
    embed(args.dataset_id, args.text_column, args.model_id, args.prefix, args.rerun, args.dimensions)

def embed(dataset_id, text_column, model_id, prefix, rerun, dimensions):
    DATA_DIR = get_data_dir()
    df = pd.read_parquet(os.path.join(DATA_DIR, dataset_id, "input.parquet"))
    
    embedding_dir = os.path.join(DATA_DIR, dataset_id, "embeddings")
    if not os.path.exists(embedding_dir):
        os.makedirs(embedding_dir)

    batch_size = 100

    # determine the embedding id
    if rerun is not None:
        embedding_id = rerun
        starting_batch = get_last_batch(os.path.join(embedding_dir, f"{embedding_id}.h5")) // batch_size
    else:
        # determine the index of the last umap run by looking in the dataset directory
        # for files named umap-<number>.json
        embedding_files = [f for f in os.listdir(embedding_dir) if re.match(r"embedding-\d+\.h5", f)]
        if len(embedding_files) > 0:
            last_umap = sorted(embedding_files)[-1]
            last_embedding_number = int(last_umap.split("-")[1].split(".")[0])
            next_embedding_number = last_embedding_number + 1
        else:
            next_embedding_number = 1
        # make the umap name from the number, zero padded to 3 digits
        embedding_id = f"embedding-{next_embedding_number:03d}"
        starting_batch = 0

    print("RUNNING:", embedding_id)
    model = get_embedding_model(model_id)
    print("loading", model.name)
    model.load_model()

    print("Checking for empty inputs")
    sentences = df[text_column].tolist()
    prefixed = []
    if prefix is None:
        prefix = ""
    for i,s in enumerate(sentences):
        if s is None or s == "":
            print(i,s, "text is empty, adding a [space]")
            s = " "
        prefixed.append(prefix + s)
    sentences = prefixed #[prefix + s for s in sentences]

    total_batches = len(sentences)//batch_size

    print("embedding", len(sentences), "sentences", "in", total_batches, "batches")
    if starting_batch > 0:
        print("Rerunning starting at batch", starting_batch)

    for i, batch in enumerate(tqdm(chunked_iterable(sentences, batch_size), total=total_batches)):
        if i < starting_batch:
            print(f"skipping batch {i}/{total_batches}", flush=True)
            continue
        try:
            embeddings = np.array(model.embed(batch, dimensions=dimensions))
            append_to_hdf5(os.path.join(embedding_dir, f"{embedding_id}.h5"), embeddings)
        except Exception as e:
            print(batch)
            print("error embedding batch", i, e)
            print("exiting prematurely", embedding_id)
            # extract the rows from the last batch from df
            df_batch = df.iloc[i*batch_size:(i+1)*batch_size].copy()
            df_batch["_ls_text_"] = batch
            batch_path = os.path.join(embedding_dir, f"{embedding_id}-batch-{i}.parquet")
            df_batch.to_parquet(batch_path)
            print("wrote original data for batch along with processed inputs in _ls_sentences_ column to\n", batch_path)
            print("debug with command:")
            print("ls-embed-debug", batch_path, model_id)

            sys.exit(1)

    with open(os.path.join(embedding_dir, f"{embedding_id}.json"), 'w') as f:
        json.dump({
            "id": embedding_id,
            "model_id": model_id,
            "dataset_id": dataset_id,
            "text_column": text_column,
            # "dimensions": np_embeds.shape[1],
            "dimensions": embeddings.shape[1],
            "prefix": prefix,
            }, f, indent=2)


    # np.save(os.path.join(embedding_dir, f"{embedding_id}.npy"), np_embeds)
    print("done with", embedding_id)

def truncate():
    parser = argparse.ArgumentParser(description='Make a copy of an existing embedding truncated to a smaller number of dimensions')
    parser.add_argument('dataset_id', type=str, help='Dataset id (directory name in data/)')
    parser.add_argument('embedding_id', type=str, help='ID of embedding to use') 
    parser.add_argument('dimensions', type=int, help='Number of dimensions to truncate to') 
    args = parser.parse_args()
    embed_truncate(args.dataset_id, args.embedding_id, args.dimensions)

def embed_truncate(dataset_id, embedding_id, dimensions):
    
    DATA_DIR = get_data_dir()
    embedding_dir = os.path.join(DATA_DIR, dataset_id, "embeddings")

    embedding_meta_path = os.path.join(embedding_dir, f"{embedding_id}.json")
    with open(embedding_meta_path, 'r') as f:
        embedding_meta = json.load(f)
    # Load the embedding model
    model_id = embedding_meta["model_id"]
    model = get_embedding_model_dict(model_id)
    print("model params", model["params"])
    # Check if the model has the attribute 'dimensions'
    try:
        dims = model["params"]['dimensions']
    except KeyError:
        raise KeyError(f"The model {model_id} does not have the 'dimensions' parameter meaning it cannot be truncated.")

    # determine the index of the last umap run by looking in the dataset directory
    # for files named umap-<number>.json
    embedding_files = [f for f in os.listdir(embedding_dir) if re.match(r"embedding-\d+\.h5", f)]
    if len(embedding_files) > 0:
        last_umap = sorted(embedding_files)[-1]
        last_embedding_number = int(last_umap.split("-")[1].split(".")[0])
        next_embedding_number = last_embedding_number + 1
    else:
        next_embedding_number = 1
    # make the umap name from the number, zero padded to 3 digits
    new_embedding_id = f"embedding-{next_embedding_number:03d}"
    print("RUNNING:", new_embedding_id)

    # read in the embeddings from embedding_id
    embedding_path = os.path.join(embedding_dir, f"{embedding_id}.h5")
    with h5py.File(embedding_path, 'r') as f:
        dataset = f["embeddings"]
        embeddings = np.array(dataset)
   

    print("truncating to", dimensions, "dimensions")
    matroyshka = embeddings[:, :dimensions]
    append_to_hdf5(os.path.join(embedding_dir, f"{new_embedding_id}.h5"), matroyshka)
    
    with open(os.path.join(embedding_dir, f"{new_embedding_id}.json"), 'w') as f:
        json.dump({
            "id": new_embedding_id,
            "model_id": embedding_meta["model_id"],
            "dataset_id": dataset_id,
            "text_column": embedding_meta["text_column"],
            "dimensions": matroyshka.shape[1],
            "prefix": embedding_meta["prefix"],
            }, f, indent=2)

    print("wrote", os.path.join(embedding_dir, f"{new_embedding_id}.h5"))
    print("done")

def debug():
    parser = argparse.ArgumentParser(description='Debug embedding a batch')
    parser.add_argument('parquet_file', type=str, help='Parquet file output by embed process')
    parser.add_argument('model_id', type=str, help='ID of embedding model to use')
    parser.add_argument('--text_column', type=str, help='Column name for text data', default="_ls_text_")
    args = parser.parse_args()
    embed_debug(args.parquet_file, args.model_id, args.text_column)

def embed_debug(parquet_file, model_id, text_column):
    df = pd.read_parquet(parquet_file)
    model = get_embedding_model(model_id)
    print("loading", model.name)
    model.load_model()

    for i,row in enumerate(df.iterrows()):
        print("batch index:", i)
        print("original index:", row[0])
        text = row[1][text_column]
        print("text:", text)
        # print("tokens:", len(model.tokenizer.encode(text)))
        # print("batch index:", i, "DataFrame index:", row[0], "Text:", row[1][text_column])
        embedding = model.embed([text])
        print("embedding", embedding)
        

if __name__ == "__main__":
   main() 


import os
import shutil
import subprocess
from setuptools import setup, find_packages
from setuptools.command.build_py import build_py

with open('requirements.txt') as f:
    required = f.read().splitlines()

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

# Build and copy the web assets
class CustomBuild(build_py):
    def run(self):
        # First run the standard build process
        super().run()

        # Run the npm build command in the web/ directory
        npm_build_command = "npm run production"
        web_dir_path = os.path.join(os.getcwd(), 'web')
        npm_build_process = subprocess.Popen(npm_build_command, cwd=web_dir_path, shell=True)
        # Ensure the npm build subprocess has finished before continuing
        npm_build_process.wait()

        # Directory containing the web assets
        web_assets_src = os.path.join(os.getcwd(), 'web/dist/production')
        # Target directory for the web assets
        web_assets_dest = os.path.join(os.getcwd(), 'latentscope/web/dist')

        # Create target directory if it doesn't exist
        os.makedirs(web_assets_dest, exist_ok=True)

        # Copy the files
        if os.path.exists(web_assets_src):
            for item in os.listdir(web_assets_src):
                s = os.path.join(web_assets_src, item)
                d = os.path.join(web_assets_dest, item)
                if os.path.isdir(s):
                    shutil.copytree(s, d, dirs_exist_ok=True)
                else:
                    shutil.copy2(s, d)

setup(
    name='latentscope',
    version='0.1.1',
    description='Quickly embed, project, cluster and explore a dataset.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/enjalot/latent-scope',
    project_urls={
        'Source': 'https://github.com/enjalot/latent-scope',
        'Tracker': 'https://github.com/enjalot/latent-scope/issues',
    },
    packages=find_packages(),
    install_requires=required,
    cmdclass={
        'build_py': CustomBuild,
    },
    entry_points={
        'console_scripts': [
            'ls-serve=latentscope.server:start',
            'ls-init=latentscope:main',
            'ls-ingest=latentscope.scripts.ingest:main',
            'ls-list-models=latentscope:list_models',
            'ls-embed=latentscope.scripts.embed:main',
            'ls-embed-debug=latentscope.scripts.embed:debug',
            'ls-embed-truncate=latentscope.scripts.embed:truncate',
            'ls-umap=latentscope.scripts.umapper:main',
            'ls-cluster=latentscope.scripts.cluster:main',
            'ls-label=latentscope.scripts.label_clusters:main',
            'ls-scope=latentscope.scripts.scope:main',
        ],
    },
    include_package_data=True,
    package_data={
        'latentscope': ['web/dist/*', 'web/dist/**/*'],
        'latentscope.models': ['embedding_models.json', 'chat_models.json'],
    },
    # rest of your setup configuration...
)



import re
import os
import sys
import json
import logging
import argparse
import pandas as pd
import pkg_resources
from flask import Flask, request, jsonify, send_from_directory
from flask_cors import CORS

# from latentscope.util import update_data_dir
from latentscope.util import get_data_dir 

app = Flask(__name__)

app.logger.addHandler(logging.StreamHandler(sys.stderr))
app.logger.setLevel(logging.INFO)

CORS(app)

# DATA_DIR = update_data_dir(args.data_dir)
DATA_DIR = get_data_dir()
if not os.path.exists(DATA_DIR):
    os.makedirs(DATA_DIR)

# We enable a read only mode of the server
def check_read_only(s):
    if s is None:
        return False
    return s.lower() in ['true', '1', 't', 'y', 'yes']
# export LATENT_SCOPE_READ_ONLY=1  
READ_ONLY = check_read_only(os.getenv("LATENT_SCOPE_READ_ONLY"))
print("READ ONLY?", READ_ONLY)

# in memory cache of dataframes loaded for each dataset
# used in returning rows for a given index (indexed, get_tags)
DATAFRAMES = {}

from .jobs import jobs_bp, jobs_write_bp
app.register_blueprint(jobs_bp, url_prefix='/api/jobs') 
if(not READ_ONLY):
    app.register_blueprint(jobs_write_bp, url_prefix='/api/jobs') 

from .search import search_bp
app.register_blueprint(search_bp, url_prefix='/api/search') 

from .tags import tags_bp, tags_write_bp
app.register_blueprint(tags_bp, url_prefix='/api/tags') 
if(not READ_ONLY):
    app.register_blueprint(tags_write_bp, url_prefix='/api/tags') 

from .datasets import datasets_bp, datasets_write_bp
app.register_blueprint(datasets_bp, url_prefix='/api/datasets')
if(not READ_ONLY):
    app.register_blueprint(datasets_write_bp, url_prefix='/api/datasets')



# ===========================================================
# File based routes for reading data and metadata from disk
# ===========================================================
@app.route('/api/embedding_models', methods=['GET'])
def get_embedding_models():
    embedding_path = pkg_resources.resource_filename('latentscope.models', 'embedding_models.json')
    with open(embedding_path, 'r', encoding='utf-8') as file:
        models = json.load(file)
    return jsonify(models)

@app.route('/api/chat_models', methods=['GET'])
def get_chat_models():
    chat_path = pkg_resources.resource_filename('latentscope.models', 'chat_models.json')
    with open(chat_path, 'r', encoding='utf-8') as file:
        models = json.load(file)
    return jsonify(models)


"""
Allow fetching of dataset files directly from disk
"""
@app.route('/api/files/<path:datasetPath>', methods=['GET'])
def send_file(datasetPath):
    print("req url", request.url)
    return send_from_directory(DATA_DIR, datasetPath)

"""
Given a list of indices (passed as a json array), return the rows from the dataset
"""
@app.route('/api/indexed', methods=['POST'])
def indexed():
    data = request.get_json()
    dataset = data['dataset']
    indices = data['indices']
    if dataset not in DATAFRAMES:
        df = pd.read_parquet(os.path.join(DATA_DIR, dataset, "input.parquet"))
        DATAFRAMES[dataset] = df
    else:
        df = DATAFRAMES[dataset]
    
    # get the indexed rows
    rows = df.iloc[indices]
    # send back the rows as json
    return rows.to_json(orient="records")
@app.route('/', defaults={'path': ''})
@app.route('/<path:path>')
def catch_all(path):
    if path != "":
        pth = pkg_resources.resource_filename('latentscope', f"web/dist/{path}")
        directory = os.path.dirname(pth)
        return send_from_directory(directory, os.path.basename(pth))
    else:
        pth = pkg_resources.resource_filename('latentscope', "web/dist/index.html")
        directory = os.path.dirname(pth)
        return send_from_directory(directory, os.path.basename(pth))

def serve(host="0.0.0.0", port=5001, debug=True):
    app.run(host=host, port=port, debug=debug)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Serve the Latent Scope API')
    # parser.add_argument('data_dir', type=str, nargs='?', default=None, help='Path to the directory where data is stored')
    parser.add_argument('--host', type=str, default="0.0.0.0", help='Host to serve the server on')
    parser.add_argument('--port', type=int, default=5001, help='Port to run the server on')
    parser.add_argument('--debug', action='store_true', help='Run server in debug mode')
    args = parser.parse_args()
    host = args.host
    port = args.port
    debug = args.debug
    serve(host, port, debug)


import os
import re
import json
import argparse
from latentscope.util import get_data_dir


def main():
    parser = argparse.ArgumentParser(description='Setup a scope')
    parser.add_argument('dataset_id', type=str, help='Dataset id (directory name in data folder)')
    parser.add_argument('embedding_id', type=str, help='Embedding id')
    parser.add_argument('umap_id', type=str, help='UMAP id')
    parser.add_argument('cluster_id', type=str, help='Cluster id')
    parser.add_argument('cluster_labels_id', type=str, help='Cluster labels id')
    parser.add_argument('label', type=str, help='Label for the scope')
    parser.add_argument('description', type=str, help='Description of the scope')

def scope(dataset_id, embedding_id, umap_id, cluster_id, cluster_labels_id, label, description):
    DATA_DIR = get_data_dir()
    print("DATA DIR", DATA_DIR)
    directory = os.path.join(DATA_DIR, dataset_id, "scopes")

    def get_next_scopes_number(dataset):
        # figure out the latest scope number
        scopes_files = [f for f in os.listdir(directory) if re.match(r"scopes-\d+\.json", f)]
        if len(scopes_files) > 0:
            last_scopes = sorted(scopes_files)[-1]
            last_scopes_number = int(last_scopes.split("-")[1].split(".")[0])
            next_scopes_number = last_scopes_number + 1
        else:
            next_scopes_number = 1
        return next_scopes_number

    next_scopes_number = get_next_scopes_number(dataset_id)
    # make the umap name from the number, zero padded to 3 digits
    id = f"scopes-{next_scopes_number:03d}"
    print("RUNNING:", id)

    scope = {
        "id": id,
        "embedding_id": embedding_id,
        "umap_id": umap_id,
        "cluster_id": cluster_id,
        "cluster_labels_id": cluster_labels_id,
        "label": label,
        "description": description
    }
    
    file_path = os.path.join(directory, id + ".json")
    with open(file_path, 'w') as f:
        json.dump(scope, f, indent=2)
    print("wrote scope", id)


import torch
from .base import EmbedModelProvider, ChatModelProvider
from transformers import AutoTokenizer, AutoModel, pipeline

def cls_pooling(model_output):
    return model_output[0][:, 0]

def average_pooling(model_output, attention_mask):
    last_hidden = model_output.last_hidden_state.masked_fill(~attention_mask[..., None].bool(), 0.0)
    return last_hidden.sum(dim=1) / attention_mask.sum(dim=1)[..., None]

def mean_pooling(model_output, attention_mask):
    token_embeddings = model_output[0]
    input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
    return torch.sum(token_embeddings * input_mask_expanded, 1) / torch.clamp(input_mask_expanded.sum(1), min=1e-9)


class TransformersEmbedProvider(EmbedModelProvider):
    def load_model(self):
        if self.name == "nomic-ai/nomic-embed-text-v1" or self.name == "nomic-ai/nomic-embed-text-v1.5":
            self.tokenizer = AutoTokenizer.from_pretrained("bert-base-uncased", model_max_length=self.params["max_tokens"])
            self.model = AutoModel.from_pretrained("nomic-ai/nomic-embed-text-v1", trust_remote_code=True, rotary_scaling_factor=2 )
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(self.name)
            self.model = AutoModel.from_pretrained(self.name, trust_remote_code=True)
        self.model.eval()

    def embed(self, inputs, dimensions=None):
        encoded_input = self.tokenizer(inputs, padding=self.params["padding"], truncation=self.params["truncation"], return_tensors='pt')
        pool = self.params["pooling"]
        # Compute token embeddings
        with torch.no_grad():
            model_output = self.model(**encoded_input)
            if pool == "cls":
                embeddings = cls_pooling(model_output)
            elif pool == "average":
                embeddings = average_pooling(model_output, encoded_input["attention_mask"])
            elif pool == "mean":
                embeddings = mean_pooling(model_output, encoded_input["attention_mask"])

        # Support Matroyshka embeddings
        if dimensions is not None and dimensions > 0:
            embeddings = torch.nn.functional.layer_norm(embeddings, normalized_shape=(embeddings.shape[1],))
            embeddings = embeddings[:, :dimensions]

        # Normalize embeddings
        normalized_embeddings = torch.nn.functional.normalize(embeddings, p=2, dim=1)
        return normalized_embeddings.tolist()


class TransformersChatProvider(ChatModelProvider):
    def load_model(self):
        # self.pipe = pipeline("text-generation", model="TinyLlama/TinyLlama-1.1B-Chat-v1.0", torch_dtype=torch.bfloat16, device_map="auto")
        # self.pipe = pipeline("text-generation", model="TinyLlama/TinyLlama-1.1B-Chat-v1.0", torch_dtype=torch.bfloat16, device_map="cpu")
        # TODO: support bfloat16 for non mac environmentss
        self.pipe = pipeline("text-generation", model="TinyLlama/TinyLlama-1.1B-Chat-v1.0", torch_dtype=torch.float16, device_map="auto")
        self.encoder = self.pipe.tokenizer

    def chat(self, messages, max_new_tokens=24):
        prompt = self.pipe.tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
        outputs = self.pipe(prompt, max_new_tokens=max_new_tokens, do_sample=True, temperature=0.7, top_k=50, top_p=0.95)
        generated_text = outputs[0]["generated_text"]
        print("GENERATED TEXT", generated_text)
        return generated_text.split("<|assistant|>")[1].strip()



from . import models
from .scripts.ingest import ingest
from .scripts.embed import embed
from .scripts.umapper import umapper as umap
from .scripts.cluster import clusterer as cluster
from .scripts.label_clusters import labeler as label
from .scripts.scope import scope

from .server import serve

from .util import update_data_dir, get_data_dir, set_openai_key, set_voyage_key, set_together_key, set_cohere_key, set_mistral_key

def init(data_dir, env_file=".env", **kwargs):
  data_dir = update_data_dir(data_dir, env_file=env_file)
  setters = {
      'openai_key': set_openai_key,
      'voyage_key': set_voyage_key,
      'together_key': set_together_key,
      'cohere_key': set_cohere_key,
      'mistral_key': set_mistral_key,
  }
  for key, setter in setters.items():
      if key in kwargs:
          setter(kwargs[key])
  print("Initialized env with data directory at", data_dir)

def main():
    import argparse
    parser = argparse.ArgumentParser(description='Initialize a data directory')
    parser.add_argument('data_dir', type=str, help='Directory to store data')
    parser.add_argument('--env_file', type=str, help='Path to .env file', default=".env")
    parser.add_argument('--openai_key', type=str, help='OpenAI API key')
    parser.add_argument('--voyage_key', type=str, help='Voyage API key')
    parser.add_argument('--together_key', type=str, help='Together API key')
    parser.add_argument('--cohere_key', type=str, help='Cohere API key')
    parser.add_argument('--mistral_key', type=str, help='Mistral API key')

    args = parser.parse_args()
    init(args.data_dir, args.env_file)

def list_models():
    ml = models.get_embedding_model_list()
    print("=== Embedding Models ===")
    for m in ml:
        # TODO: pretty print the relevant model info
        print(m["id"])
    print("\n")
    print("=== Summarization (Chat) Models ===")
    ml = models.get_chat_model_list()
    for m in ml:
        print(m["id"])


import os
import time
import cohere
from .base import EmbedModelProvider

from latentscope.util import get_key

class CohereAIEmbedProvider(EmbedModelProvider):
    def load_model(self):
        api_key = get_key("COHERE_API_KEY")
        if api_key is None:
            print("ERROR: No API key found for Cohere")
            print("Missing 'COHERE_API_KEY' variable in:", f"{os.getcwd()}/.env")
        self.client = cohere.Client(api_key)

    def embed(self, inputs):
        time.sleep(0.01) # TODO proper rate limiting
        response = self.client.embed(texts=inputs, model=self.name, input_type=self.params["input_type"])
        embeddings = response.embeddings
        return embeddings


import os
import json
import h5py
import pandas as pd
import numpy as np
from flask import Blueprint, jsonify, request
from sklearn.neighbors import NearestNeighbors

from latentscope.models import get_embedding_model

# Create a Blueprint
search_bp = Blueprint('search_bp', __name__)
DATA_DIR = os.getenv('LATENT_SCOPE_DATA')

# in memory cache of dataset metadata, embeddings, models and tokenizers
DATASETS = {}
EMBEDDINGS = {}

"""
Returns nearest neighbors for a given query string
Hard coded to 150 results currently
"""
@search_bp.route('/nn', methods=['GET'])
def nn():
    dataset = request.args.get('dataset')
    embedding_id = request.args.get('embedding_id')
    dimensions = request.args.get('dimensions')
    dimensions = int(dimensions) if dimensions else None
    print("dimensions", dimensions)

    num = 150
    if embedding_id not in EMBEDDINGS:
        print("loading model", embedding_id)
        with open(os.path.join(DATA_DIR, dataset, "embeddings", embedding_id + ".json"), 'r') as f:
            metadata = json.load(f)
        model_id = metadata.get('model_id')
        print("Model ID:", model_id)
        model = get_embedding_model(model_id)
        model.load_model()
        EMBEDDINGS[embedding_id] = model
    else:
        model = EMBEDDINGS[embedding_id]

    if dataset not in DATASETS or embedding_id not in DATASETS[dataset]:
        # load the dataset embeddings
        # embeddings = np.load(os.path.join(DATA_DIR, dataset, "embeddings", embedding_id + ".npy"))
        embedding_path = os.path.join(DATA_DIR, dataset, "embeddings", f"{embedding_id}.h5")
        with h5py.File(embedding_path, 'r') as f:
            embeddings = np.array(f["embeddings"])
        print("fitting embeddings")
        nne = NearestNeighbors(n_neighbors=num, metric="cosine")
        nne.fit(embeddings)
        if dataset not in DATASETS:
          DATASETS[dataset] = {}
        DATASETS[dataset][embedding_id] = nne
    else:
        nne = DATASETS[dataset][embedding_id]
    
    # embed the query string and find the nearest neighbor
    query = request.args.get('query')
    print("query", query)
    embedding = model.embed([query], dimensions=dimensions)
    distances, indices = nne.kneighbors(embedding)
    # print("distances", distances)
    # Filter distances and indices to only elements where distance is less than .4
    # filtered_indices = indices[0][distances[0] < 0.4]
    # filtered_distances = distances[0][distances[0] < 0.4]
    filtered_indices = indices[0]
    filtered_distances = distances[0]
    indices = filtered_indices
    distances = filtered_distances
        
    return jsonify(indices=indices.tolist(), distances=distances.tolist())


@search_bp.route('/compare', methods=['GET'])
def compare():
    dataset = request.args.get('dataset')
    umap_left = request.args.get('umap_left')
    umap_right = request.args.get('umap_right')
    k = request.args.get('k')
    k = int(k) if k else 5

    umap_dir = os.path.join(DATA_DIR, dataset, "umaps")
    left_df = pd.read_parquet(os.path.join(umap_dir, f"{umap_left}.parquet"))
    left = left_df.to_numpy()
    right_df = pd.read_parquet(os.path.join(umap_dir, f"{umap_right}.parquet"))
    right = right_df.to_numpy()

    # Calculate the absolute displacement
    absolute_displacement = np.linalg.norm(right - left, axis=1)
    min_abs_displacement = np.min(absolute_displacement)
    max_abs_displacement = np.max(absolute_displacement)
    if max_abs_displacement - min_abs_displacement > 0:
        absolute_displacement = (absolute_displacement - min_abs_displacement) / (max_abs_displacement - min_abs_displacement)
    else:
        absolute_displacement = np.zeros_like(absolute_displacement)


    # Compute nearest neighbors in both projections
    # knn = NearestNeighbors(n_neighbors=k+1, metric="euclidean")  # +1 because the point itself is included
    # knn.fit(left)
    # distances1, indices1 = knn.kneighbors(left)

    # knn.fit(right)
    # distances2, indices2 = knn.kneighbors(right)

    # relative_displacement = np.abs(distances1 - distances2).mean(axis=1)
    # # Normalize relative_displacement
    # min_relative_displacement = np.min(relative_displacement)
    # max_relative_displacement = np.max(relative_displacement)
    # if max_relative_displacement - min_relative_displacement > 0:
    #     relative_displacement = (relative_displacement - min_relative_displacement) / (max_relative_displacement - min_relative_displacement)
    # else:
    #     relative_displacement = np.zeros_like(relative_displacement)

    
    # size = left.shape[0]
    # # Calculate displacement scores
    # displacement_scores = np.zeros(size)
    # for i in range(size):
    #     # Find the actual positions (0 to k-1) of common neighbors in each list
    #     neighbor_positions_1 = {index: pos for pos, index in enumerate(indices1[i]) if index in indices2[i]}
    #     neighbor_positions_2 = {index: pos for pos, index in enumerate(indices2[i]) if index in indices1[i]}
        
    #     # Calculate displacement for common neighbors
    #     displacements = []
    #     for index, pos1 in neighbor_positions_1.items():
    #         pos2 = neighbor_positions_2[index]
    #         displacement = abs(distances1[i, pos1] - distances2[i, pos2])
    #         displacements.append(displacement)
        
    #     # Compute the mean displacement for the point, if there are common neighbors
    #     if displacements:
    #         displacement_scores[i] = np.mean(displacements)

    # # normalize displacement_scores from 0 to 1
    # min_score = np.min(displacement_scores)
    # max_score = np.max(displacement_scores)
    # if max_score - min_score > 0:
    #     displacement_scores = (displacement_scores - min_score) / (max_score - min_score)
    # else:
    #     displacement_scores = np.zeros_like(displacement_scores)

    # TODO: why don't these actually add up to 1 even when i normalize each
    # combined_scores = 0.6 * absolute_displacement + 0.6 * relative_displacement + 0.6 * displacement_scores
    # combined_scores = (absolute_displacement + relative_displacement + displacement_scores) / 3
    combined_scores = absolute_displacement #(absolute_displacement + relative_displacement + displacement_scores) / 3
    return jsonify(combined_scores.tolist())
    # return jsonify(displacement_scores.tolist())


